use anyhow::Error;
use gc_arena::{Gc, Rootable};
use piccolo::{
    self, Callback, Closure, Context, Executor, FromValue, Function, IntoValue, Lua,
    StashedExecutor, StaticError, Table, UserData, Value, Variadic,
};
use wasm_bindgen::JsValue;
use web_sys::console;
use yew::UseStateSetter;

use crate::{
    command_handler::debrace,
    echo_to_term_frame,
    id_intern::intern_id,
    websocket::{connect_websocket, send_message_to_mud, WebSocketId},
    GlobalLayoutCell, GlobalLayoutState, GlobalMemoCell, TermFrame,
};
use std::{collections::VecDeque, rc::Rc, str};

pub struct LuaState {
    pub interp: Lua,
    pub exec: StashedExecutor,
}

impl LuaState {
    pub fn setup() -> Result<LuaState, String> {
        let mut interp = Lua::core();
        let exec: StashedExecutor =
            interp.enter(|ctx| Ok::<StashedExecutor, String>(ctx.stash(Executor::new(ctx))))?;

        Ok(LuaState { interp, exec })
    }

    fn try_set_current_frame(&mut self, frame: &TermFrame) -> Result<(), StaticError> {
        self.interp.try_enter(|ctx| {
            let info_table = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"info")))?;
            info_table.set(
                ctx,
                ctx.intern_static(b"current_frame"),
                intern_id::<TermFrame>(ctx, frame.clone()),
            )?;
            Ok(())
        })
    }

    pub fn set_current_frame(&mut self, frame: &TermFrame) {
        // We silently ignore errors here. Failure can happen if the Lua code does weird things
        // like messes with the info global, so better to just ignore.
        self.try_set_current_frame(frame).unwrap_or(())
    }

    pub fn execute(&mut self, command: &str) -> Result<(), String> {
        self.interp
            .try_enter(|ctx| {
                let closure = Closure::load(ctx, None, format!("return ({})", command).as_bytes())
                    .or_else(|_| Closure::load(ctx, None, command.as_bytes()))?;
                ctx.fetch(&self.exec).restart(ctx, closure.into(), ());
                Ok(())
            })
            .map_err(|err| format!("{}", err))?;
        self.interp
            .execute::<()>(&self.exec)
            .map_err(|err| format!("{}", err))
    }

    pub fn execute_command(
        &mut self,
        command: &str,
        arguments: &VecDeque<&str>,
    ) -> Result<(), String> {
        self.interp
            .try_enter(|ctx| {
                let commands =
                    Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"commands")))?;
                let command_value: Value = commands.get(ctx, ctx.intern(command.as_bytes()));
                if command_value.is_nil() {
                    Err(anyhow::Error::msg("Unknown command"))?;
                }
                let command_fn = Function::from_value(ctx, command_value)?;
                ctx.fetch(&self.exec).restart(
                    ctx,
                    command_fn,
                    Variadic(
                        arguments
                            .iter()
                            .map(|s| ctx.intern(s.as_bytes()).into())
                            .collect::<Vec<Value>>(),
                    ),
                );
                Ok(())
            })
            .map_err(|err| format!("{}", err))?;
        self.interp
            .execute::<()>(&self.exec)
            .map_err(|err| format!("{}", err))
    }
}

pub fn install_lua_globals(
    global_memo: &GlobalMemoCell,
    global_layout: UseStateSetter<GlobalLayoutCell>,
) -> Result<(), String> {
    global_memo
        .lua_engine
        .borrow_mut()
        .interp
        .try_enter(|ctx| {
            let cmd_table = Table::new(&ctx);
            macro_rules! register_command {
                ($sym: ident) => {
                    cmd_table
                        .set(
                            ctx,
                            ctx.intern_static(stringify!($sym).as_bytes()),
                            $sym(ctx, &global_memo, &global_layout),
                        )
                        .map_err(|_| Error::msg("Can't add command"))?;
                };
            }

            register_command!(echo);
            register_command!(echo_frame);
            register_command!(echo_frame_raw);
            register_command!(connect_mud);
            register_command!(sendmud_raw);
            register_command!(hsplit);
            register_command!(panel_merge);
            register_command!(vsplit);
            ctx.set_global(ctx.intern_static(b"commands").into_value(ctx), cmd_table)
                .map(|_| ())
                .map_err(|_| Error::msg("Can't set commands key"))?;

            let info_table = Table::new(&ctx);
            ctx.set_global(ctx.intern_static(b"info").into_value(ctx), info_table)
                .map(|_| ())
                .map_err(|_| Error::msg("Can't set info key"))?;
            let muds_table = Table::new(&ctx);
            ctx.set_global(ctx.intern_static(b"muds").into_value(ctx), muds_table)
                .map(|_| ())
                .map_err(|_| Error::msg("Can't set muds key"))?;

            Ok(())
        })
        .map_err(|e| e.to_string())?;

    Ok(())
}

fn echo_frame_raw<'gc, 'a>(
    ctx: Context<'gc>,
    global_memo: &'a GlobalMemoCell,
    _global_layout: &'a UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let frame: TermFrame = try_unwrap_frame(ctx, &stack.pop_front())?;
        let message: piccolo::String = stack.from_front(ctx)?;
        let message_str = str::from_utf8(message.as_bytes())
            .map_err(|_| "Expected message to echo to be UTF-8.".into_value(ctx))?;
        echo_to_term_frame(&global_memo, &frame, message_str).map_err(|m| m.into_value(ctx))?;
        Ok(piccolo::CallbackReturn::Return)
    })
}

fn echo_frame<'gc>(
    ctx: Context<'gc>,
    _global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let commands: Table<'gc> =
            Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"commands")))?;
        let function = Function::from_value(ctx, commands.get(ctx, "echo_frame_raw"))?;
        let frame_no: Value = stack.pop_front();
        let all_parts: Vec<String> = stack
            .consume::<Variadic<Vec<Value>>>(ctx)?
            .into_iter()
            .map(|v| format!("{}", v))
            .collect();
        stack.push_front(frame_no);
        let message = ctx.intern((all_parts.join(" ") + "\r\n").as_bytes());
        stack.push_back(message.into());
        Ok(piccolo::CallbackReturn::Call {
            function,
            then: None,
        })
    })
}

fn echo<'gc>(
    ctx: Context<'gc>,
    _global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let commands: Table<'gc> =
            Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"commands")))?;
        let function = Function::from_value(ctx, commands.get(ctx, "echo_frame"))?;
        let info: Table<'gc> = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"info")))?;
        let cur_frame = info.get(ctx, ctx.intern_static(b"current_frame"));
        stack.push_front(cur_frame);
        Ok(piccolo::CallbackReturn::Call {
            function,
            then: None,
        })
    })
}

fn vsplit<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let path: &str = debrace(&path);
        let frame: u64 = stack.from_front(ctx)?;
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .vsplit(path, TermFrame(frame))
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(piccolo::CallbackReturn::Return)
    })
}

fn hsplit<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let path: &str = debrace(&path);
        let frame: u64 = stack.from_front(ctx)?;
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .hsplit(path, TermFrame(frame))
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(piccolo::CallbackReturn::Return)
    })
}

fn panel_merge<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let path: &str = debrace(&path);
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .join(path)
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(piccolo::CallbackReturn::Return)
    })
}

fn try_unwrap_frame<'gc>(
    ctx: Context<'gc>,
    value: &Value<'gc>,
) -> Result<TermFrame, piccolo::Error<'gc>> {
    match u64::from_value(ctx, *value) {
        Ok(v) => Ok(TermFrame(v)),
        Err(_) => Ok(UserData::from_value(ctx, *value)?
            .downcast::<Rootable!['gcb => Gc<'gcb, TermFrame>]>()?
            .as_ref()
            .clone()),
    }
}

fn try_unwrap_socketid<'gc>(
    ctx: Context<'gc>,
    value: &Value<'gc>,
) -> Result<WebSocketId, piccolo::Error<'gc>> {
    Ok(UserData::from_value(ctx, *value)?
        .downcast::<Rootable!['gcb => Gc<'gcb, WebSocketId>]>()?
        .as_ref()
        .clone())
}

pub fn handle_websocket_output(
    socket: &WebSocketId,
    engine: &mut LuaState,
    input: &[u8],
) -> Result<(), String> {
    engine
        .interp
        .try_enter(|ctx| {
            let handlers = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"handlers")))?;
            let input_fn =
                Function::from_value(ctx, handlers.get(ctx, ctx.intern_static(b"mudoutput")))?;
            ctx.fetch(&engine.exec).restart(
                ctx,
                input_fn,
                (intern_id(ctx, socket.clone()), ctx.intern(input)),
            );
            Ok(())
        })
        .map_err(|err| format!("{}", err))?;
    engine
        .interp
        .execute::<()>(&engine.exec)
        .map_err(|err| format!("{}", err))
}

pub fn handle_websocket_output_log_err(socket: &WebSocketId, engine: &mut LuaState, input: &[u8]) {
    match handle_websocket_output(socket, engine, input) {
        Ok(()) => {}
        Err(e) => console::log_2(
            &JsValue::from_str("An error occurred calling the WebSocket input handler"),
            &JsValue::from_str(&e),
        ),
    }
}

pub fn handle_websocket_close(socket: &WebSocketId, engine: &mut LuaState) -> Result<(), String> {
    engine
        .interp
        .try_enter(|ctx| {
            let handlers = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"handlers")))?;
            let input_fn =
                Function::from_value(ctx, handlers.get(ctx, ctx.intern_static(b"mudclose")))?;
            ctx.fetch(&engine.exec)
                .restart(ctx, input_fn, intern_id(ctx, socket.clone()));
            Ok(())
        })
        .map_err(|err| format!("{}", err))?;
    engine
        .interp
        .execute::<()>(&engine.exec)
        .map_err(|err| format!("{}", err))
}

pub fn handle_websocket_close_log_err(socket: &WebSocketId, engine: &mut LuaState) {
    match handle_websocket_close(socket, engine) {
        Ok(()) => {}
        Err(e) => console::log_2(
            &JsValue::from_str("An error occurred calling the WebSocket input handler"),
            &JsValue::from_str(&e),
        ),
    }
}

fn sendmud_raw<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let mud: WebSocketId = try_unwrap_socketid(ctx, &stack.pop_front())?;
        let msg: piccolo::String = stack.from_front(ctx)?;
        send_message_to_mud(&mud, msg.as_bytes(), &global_memo)?;
        Ok(piccolo::CallbackReturn::Return)
    })
}

fn connect_mud<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let mut trusted: bool = false;
        let name: String = loop {
            let v: String = stack.from_front(ctx)?;
            if v == "-trust" {
                trusted = true;
                continue;
            }
            break v;
        };
        let name: Value<'gc> = ctx.intern(debrace(&name).as_bytes()).into();

        let muds = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"muds")))?;
        if !muds.get_value(name).is_nil() {
            Err(Error::msg(
                "Attempt to create MUD connection using name that's already taken",
            ))?
        }

        let url: String = stack.from_front(ctx)?;
        let new_socket = intern_id(
            ctx,
            connect_websocket(
                trusted,
                &url,
                &mut global_memo.ws_registry.borrow_mut(),
                &global_memo,
            )?,
        );

        muds.set(ctx, name, new_socket)?;
        Ok(piccolo::CallbackReturn::Return)
    })
}
